package array_list

import (
	"errors"
	"fmt"
	"slices"

	"git.tswf.io/incredible-go/incredible-go-core/pkg/container/optional"
)

func (self *ArrayList[T]) Add(element T) {
	if len(self.content) <= self.size {
		self.content = append(self.content, element)
	} else {
		self.content[self.size] = element
	}

	self.size++
}

func (self *ArrayList[T]) RemoveIf(filter func(T) bool) {
	index := 0
	for {
		if index >= self.size {
			break
		}

		element := self.content[index]
		shouldRemove := filter(element)

		if shouldRemove {
			self.RemoveByIndex(index)
		} else {
			index++
		}
	}
}

func (self *ArrayList[T]) Remove(element T) {
	indexFromMove := -1
	for index, value := range self.content {

		if any(value) == any(element) {
			indexFromMove = index
			break
		}
	}

	self.RemoveByIndex(indexFromMove)
}

func (self *ArrayList[T]) RemoveByIndex(index int) {
	contentSize := len(self.content)

	if index >= 0 {
		for i := index + 1; i < contentSize; i++ {
			self.content[i-1] = self.content[i]
		}

		self.size--
	}
}

func (self *ArrayList[T]) Get(index int) T {
	if index < 0 {
		panic(any(errors.New(fmt.Sprintf("Negative list index %d", index))))
	}

	if index >= self.size {
		panic(any(errors.New(fmt.Sprintf("Index %d bigger (or equal?) than list size %d", index, self.size))))
	}

	return self.content[index]
}

func (self *ArrayList[T]) IndexOf(value T) int {
	if self.size == 0 {
		return -1
	}

	for index, val := range self.content {
		if any(val) == any(value) {
			return index
		}
	}

	return -1
}

func (self *ArrayList[T]) Contains(value T) bool {
	return self.IndexOf(value) >= 0
}

func (self *ArrayList[T]) None(test func(T) bool) bool {
	return !self.Any(test)
}

func (self *ArrayList[T]) AddAll(list *ArrayList[T]) {
	list.ForEach(func(t T) {
		self.Add(t)
	})
}

func (self *ArrayList[T]) AddAllS(slice *[]T) {
	for _, v := range *slice {
		self.Add(v)
	}
}

func (self *ArrayList[T]) SetAll(list *ArrayList[T]) {
	self.Clear()
	self.AddAll(list)
}

func (self *ArrayList[T]) SetAllS(list *[]T) {
	self.Clear()
	self.AddAllS(list)
}

func (self *ArrayList[T]) Any(test func(T) bool) bool {
	for index, value := range self.content {
		if index >= self.size {
			return false
		}

		if test(value) {
			return true
		}
	}

	return false
}

func (self *ArrayList[T]) ForEach(fun func(T)) {
	for index, value := range self.content {
		if index >= self.size {
			return
		}

		fun(value)
	}
}

func (self *ArrayList[T]) ForEachBreakable(fun func(T, *bool)) {
	needsBreak := false
	for index, value := range self.content {
		if index >= self.size {
			return
		}

		if needsBreak {
			break
		}

		fun(value, &needsBreak)
	}
}

func (self *ArrayList[T]) Size() int {
	return self.size
}

func (self *ArrayList[T]) IsEmpty() bool {
	return self.Size() <= 0
}

func (self *ArrayList[T]) String() string {
	return fmt.Sprintf("%v", self.content)
}

func (self *ArrayList[T]) Clear() {
	self.content = self.content[:0]
}

func (self *ArrayList[T]) GetLast() T {
	if self.IsEmpty() {
		panic(any(errors.New("trying to get last element from empty array list")))
	}

	return self.Get(self.Size() - 1)
}

func (self *ArrayList[T]) GetFirst() T {
	if self.IsEmpty() {
		panic(any(errors.New("trying to get first element from empty array list")))
	}

	return self.Get(0)
}

func (self *ArrayList[T]) ToSlice() []T {
	result := slices.Clone(self.content)[:self.size]
	return result
}

func (self *ArrayList[T]) ToSliceUnsafe() []T {
	return self.content[:self.size]
}

func (self *ArrayList[T]) Find(filter func(T) bool) *optional.Optional[T] {
	for index, value := range self.content {
		if index >= self.size {
			break
		}

		passed := filter(value)

		if passed {
			return optional.OfAny(value)
		}
	}

	return optional.Empty[T]()
}

func (self *ArrayList[T]) Filter(filter func(T) bool) *ArrayList[T] {
	return self.FilterTo(filter, NewArrayList[T]())
}

func (self *ArrayList[T]) FilterTo(filter func(T) bool, resultList *ArrayList[T]) *ArrayList[T] {
	if resultList == nil {
		panic(fmt.Errorf("can not filter %v to nil list", self))
	}

	self.ForEach(func(value T) {
		if filter(value) {
			resultList.Add(value)
		}
	})

	return resultList
}
